import os
import nbrmd
import nbformat


def update_rmd_and_ipynb(model, path, format=['.ipynb', '.Rmd'], **kwargs):
    """
    A pre-save hook for jupyter that saves the notebooks under the alternative form.
    When the notebook has extension '.ipynb', this creates a '.Rmd' file
    When the notebook has extension '.Rmd', this creates a '.ipynb' file
    :param model: data model, that may contain the notebook
    :param path: full name for ipython notebook
    :param format: list of alternative formats
    :param kwargs: not used
    :return:
    """

    # only run on notebooks
    if model['type'] != 'notebook':
        return

    # only run on nbformat v4
    nb = model['content']
    if nb['nbformat'] != 4:
        return

    format = nb.get('metadata', {}).get('nbrmd_formats', format)
    if not isinstance(format, list) or not set(format).issubset(['.Rmd', '.md', '.ipynb']):
        raise TypeError("Notebook metadata 'nbrmd_formats' should be subset of ['.Rmd', '.md', '.ipynb']")

    file, ext = os.path.splitext(path)

    for alt_ext in format:
        if ext != alt_ext:
            nbrmd.writef(nbformat.notebooknode.from_dict(nb), file + alt_ext)


def update_rmd(model, path, **kwargs):
    """
    A pre-save hook for jupyter that saves the notebooks in '.Rmd' format when
    the notebook has extension '.ipynb'
    :param model: data model, that may contain the notebook
    :param path: full name for ipython notebook
    :param kwargs: not used
    :return:
    """
    update_rmd_and_ipynb(model, path, format=['.Rmd'], **kwargs)


def update_ipynb(model, path, **kwargs):
    """
    A pre-save hook for jupyter that saves the notebooks in '.Rmd' format when
    the notebook has extension '.ipynb'
    :param model: data model, that may contain the notebook
    :param path: full name for ipython notebook
    :param kwargs: not used
    :return:
    """
    update_rmd_and_ipynb(model, path, format=['.ipynb'], **kwargs)


def update_selected_formats(model, path, **kwargs):
    """
    A pre-save hook for jupyter that saves the notebooks in the formats selected in
    notebook metadata 'nbrmd_formats', that should be a list
    :param model: data model, that may contain the notebook
    :param path: full name for ipython notebook
    :param kwargs: not used
    :return:
    """

    update_rmd_and_ipynb(model, path, format=[], **kwargs)



"""
Convert between R markdown chunk options and jupyter cell metadata.

metadata.hide_input and metadata.hide_output are documented here:
http://jupyter-contrib-nbextensions.readthedocs.io/en/latest/nbextensions/runtools/readme.html

TODO: Update this if a standard gets defined at https://github.com/jupyter/notebook/issues/3700
"""

_boolean_options_dictionary = [('hide_input', 'echo', True), ('hide_output', 'include', True)]
_ignore_metadata = ['collapsed', 'autoscroll', 'deletable', 'format', 'trusted', 'tags']


def _r_logical_values(bool):
    return 'TRUE' if bool else 'FALSE'


class RLogicalValueError(Exception):
    pass


def _py_logical_values(rbool):
    if rbool in ['TRUE', 'T']:
        return True
    if rbool in ['FALSE', 'F']:
        return False
    raise RLogicalValueError


def to_chunk_options(language, metadata):
    options = language.lower()
    if 'name' in metadata:
        options += ' ' + metadata['name'] + ','
        del metadata['name']
    for jo, ro, rev in _boolean_options_dictionary:
        if jo in metadata:
            options += ' {}={},'.format(ro, _r_logical_values(metadata[jo] != rev))
            del metadata[jo]
    for co_name in metadata:
        co_value = metadata[co_name]
        co_name = co_name.strip()
        if co_name in _ignore_metadata:
            continue
        if co_value is None:
            options += ' {},'.format(co_name)
        elif isinstance(co_value, str):
            options += ' {}={},'.format(co_name, co_value)
        elif isinstance(co_value, bool):
            options += ' {}={},'.format(co_name, 'TRUE' if co_value else 'FALSE')
        else:
            options += ' {}="{}",'.format(co_name, str(co_value))
    return options.strip(',')


def update_metadata_using_dictionary(name, value, metadata):
    for jo, ro, rev in _boolean_options_dictionary:
        if name == ro:
            try:
                metadata[jo] = _py_logical_values(value) != rev
                return True
            except RLogicalValueError:
                pass
    return False


def to_metadata(options):
    options = options.split(' ', 1)
    if len(options) == 1:
        language = options[0]
        chunk_options = []
    else:
        language, others = options
        chunk_options = others.split(',')
    language = 'R' if language == 'r' else language
    metadata = {}
    for i, co in enumerate(chunk_options):
        co = co.split('=', 1)
        if len(co) == 1:
            name = co[0].strip(' ')
            if i == 0:
                metadata['name'] = name
                continue
            metadata[name] = None
        else:
            name, value = co
            name = name.strip()

            if update_metadata_using_dictionary(name, value, metadata):
                continue
            try:
                metadata[name] = _py_logical_values(value)
                continue
            except RLogicalValueError:
                metadata[name] = value
    return language, metadata



import notebook.transutils
from notebook.services.contents.filemanager import FileContentsManager
from .hooks import update_selected_formats

import os
import nbrmd
import nbformat
import mock


def _nbrmd_writes(nb, version=nbformat.NO_CONVERT, **kwargs):
    return nbrmd.writes(nb, **kwargs)


def _nbrmd_reads(s, as_version, **kwargs):
    return nbrmd.reads(s, **kwargs)


def _nbrmd_md_writes(nb, version=nbformat.NO_CONVERT, **kwargs):
    return nbrmd.nbrmd.md_writes(nb, **kwargs)


def _nbrmd_md_reads(s, as_version, **kwargs):
    return nbrmd.nbrmd.md_reads(s, **kwargs)


class RmdFileContentsManager(FileContentsManager):
    """
    A FileContentsManager Class that reads and stores notebooks to classical
    Jupyter notebooks (.ipynb), or in Rmarkdown format (.Rmd), in either .Rmd or .md files
    """
    nb_extensions = ['.ipynb', '.Rmd', '.md']

    def __init__(self, **kwargs):
        self.pre_save_hook = update_selected_formats
        super(RmdFileContentsManager, self).__init__(**kwargs)

    def _read_notebook(self, os_path, as_version=4):
        """Read a notebook from an os path."""
        file, ext = os.path.splitext(os_path)
        if ext == '.Rmd':
            with mock.patch('nbformat.reads', _nbrmd_reads):
                return super(RmdFileContentsManager, self)._read_notebook(os_path, as_version)
        elif ext == '.md':
            with mock.patch('nbformat.reads', _nbrmd_md_reads):
                return super(RmdFileContentsManager, self)._read_notebook(os_path, as_version)
        else:
            return super(RmdFileContentsManager, self)._read_notebook(os_path, as_version)

    def _save_notebook(self, os_path, nb):
        """Save a notebook to an os_path."""
        file, ext = os.path.splitext(os_path)
        if ext == '.Rmd':
            with mock.patch('nbformat.writes', _nbrmd_writes):
                return super(RmdFileContentsManager, self)._save_notebook(os_path, nb)
        elif ext == '.md':
            with mock.patch('nbformat.writes', _nbrmd_md_writes):
                return super(RmdFileContentsManager, self)._save_notebook(os_path, nb)
        else:
            return super(RmdFileContentsManager, self)._save_notebook(os_path, nb)

    def get(self, path, content=True, type=None, format=None):
        """ Takes a path for an entity and returns its model

        Parameters
        ----------
        path : str
            the API path that describes the relative path for the target
        content : bool
            Whether to include the contents in the reply
        type : str, optional
            The requested type - 'file', 'notebook', or 'directory'.
            Will raise HTTPError 400 if the content doesn't match.
        format : str, optional
            The requested format for file contents. 'text' or 'base64'.
            Ignored if this returns a notebook or directory model.

        Returns
        -------
        model : dict
            the contents model. If content=True, returns the contents
            of the file or directory as well.
        """
        path = path.strip('/')

        if self.exists(path) and \
                (type == 'notebook' or
                 (type is None and
                  any([path.endswith(ext) for ext in self.nb_extensions]))):
            return self._notebook_model(path, content=content)
        else:
            return super(RmdFileContentsManager, self).get(path, content, type, format)

    def rename_file(self, old_path, new_path):
        old_file, org_ext = os.path.splitext(old_path)
        new_file, new_ext = os.path.splitext(new_path)
        if org_ext in self.nb_extensions and org_ext == new_ext:
            for ext in self.nb_extensions:
                if self.file_exists(old_file + ext):
                    super(RmdFileContentsManager, self).rename_file(old_file + ext, new_file + ext)
        else:
            super(RmdFileContentsManager, self).rename_file(old_path, new_path)



"""R markdown notebook format for Jupyter

Use this module to read or write Jupyter notebooks as Rmd documents (methods 'read', 'reads', 'write', 'writes')

Use the 'pre_save_hook' method (see its documentation) to automatically dump your Jupyter notebooks as a Rmd file, in addition
to the ipynb file.

Use the 'nbrmd' conversion script to convert Jupyter notebooks from/to R markdown notebooks.
"""

from .nbrmd import read, reads, readf, write, writes, writef
from .hooks import update_rmd, update_ipynb, update_rmd_and_ipynb, update_selected_formats
try:
  from .cm import RmdFileContentsManager
except ImportError as e:
  RmdFileContentsManager = str(e)



"""Read and write notebooks as RStudio notebook files, with .Rmd extension.

Raw and markdown cells are converted to markdown, while code cells are converted
to code chunks. The transformation is reversible and all inputs are preserved (not
outputs, though).

Authors:

* Marc Wouts
"""

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------

import os
import re
from enum import Enum
from nbformat.v4.rwbase import NotebookReader, NotebookWriter
from nbformat.v4.nbbase import (
    new_code_cell, new_markdown_cell, new_raw_cell, new_notebook
)
import nbformat
import yaml

from .chunk_options import to_metadata, to_chunk_options


# -----------------------------------------------------------------------------
# Code
# -----------------------------------------------------------------------------

def _language(metadata):
    try:
        return metadata.language_info.name.lower()
    except AttributeError:
        return u'python'


class State(Enum):
    NONE = 0
    HEADER = 1
    MARKDOWN = 2
    CODE = 3


_header_re = re.compile(r"^---\s*")
_end_code_re = re.compile(r"^```\s*")
_r_cell_re = re.compile(r"^%%R\s*")

_jupyter_languages = ['R', 'bash', 'sh', 'python', 'python2', 'python3', 'javascript', 'js', 'perl']
_jupyter_languages_re = [re.compile(r"^%%{}\s*".format(lang)) for lang in _jupyter_languages]


class RmdReaderError(Exception):
    pass


class RmdReader(NotebookReader):

    def __init__(self, markdown=False):
        self.start_code_re = re.compile(r"^```(.*)\s*") if markdown else re.compile(r"^```\{(.*)\}\s*")

    def reads(self, s, **kwargs):
        return self.to_notebook(s, **kwargs)

    def to_notebook(self, s, **kwargs):
        lines = s.splitlines()

        metadata = {}
        cells = []
        cell_lines = []

        def add_markdown_cell():
            if len(cell_lines) == 0:
                return

            if cell_lines[-1] == '':
                if len(cell_lines) > 1 or len(cells) == 0:
                    cells.append(new_markdown_cell(source=u'\n'.join(cell_lines[:-1])))
                else:
                    cells[-1]['metadata']['noskipline'] = True
                    cells.append(new_markdown_cell(source=u'\n'.join(cell_lines)))
            else:
                cells.append(new_markdown_cell(source=u'\n'.join(cell_lines),
                                               metadata={u'noskipline': True}))

        cell_metadata = {}
        state = State.NONE
        testblankline = False

        for line in lines:
            if state is State.NONE:
                if _header_re.match(line):
                    state = State.HEADER
                    continue
                state = State.MARKDOWN

            if state is State.HEADER:
                if _header_re.match(line):
                    header = []
                    jupyter = []
                    in_header = True
                    for l in cell_lines:
                        if l.rstrip() == 'jupyter:':
                            in_header = False
                        if in_header:
                            header.append(l)
                        else:
                            jupyter.append(l)
                    if len(header):
                        cells.append(new_raw_cell(source=u'\n'.join(['---'] + header + ['---'])))
                    if len(jupyter):
                        metadata = yaml.load(u'\n'.join(jupyter))['jupyter']
                    cell_lines = []
                    state = State.MARKDOWN
                    testblankline = True
                    continue

            if testblankline:
                # Set 'noskipline' metadata if no blank line is found after cell
                testblankline = False
                if line == u'':
                    continue
                else:
                    if len(cells):
                        cells[-1][u'metadata'][u'noskipline'] = True

            if state is State.MARKDOWN:
                if self.start_code_re.match(line):
                    add_markdown_cell()
                    cell_lines = []

                    chunk_options = self.start_code_re.findall(line)[0]
                    language, cell_metadata = to_metadata(chunk_options)
                    cell_metadata['language'] = language
                    state = State.CODE
                    continue

            if state is State.CODE:
                if _end_code_re.match(line):
                    cells.append(new_code_cell(source=u'\n'.join(cell_lines),
                                               metadata=cell_metadata))
                    cell_lines = []
                    cell_metadata = {}
                    state = State.MARKDOWN
                    testblankline = True
                    continue

            cell_lines.append(line)

        # Append last cell if not empty
        if state is State.MARKDOWN:
            add_markdown_cell()
        elif state is State.CODE:
            raise RmdReaderError(u'Unterminated code cell')
        elif state is State.HEADER:
            raise RmdReaderError(u'Unterminated YAML header')

        # Determine main language
        main_language = metadata.get('language_info', {}).get('name')
        if main_language is None:
            languages = dict(python=0)
            for c in cells:
                if c['cell_type'] == 'code':
                    language = c['metadata']['language']
                    if language == 'r':
                        language = 'R'
                    languages[language] = languages.get(language, 0) + 1

            main_language = max(languages, key=languages.get)
            metadata['language_info'] = metadata.get('language_info', {})
            metadata['language_info']['name'] = main_language

        # Remove 'language' meta data and add a magic if not main language
        for c in cells:
            if c['cell_type'] == 'code':
                language = c['metadata']['language']
                del c['metadata']['language']
                if language != main_language and language in _jupyter_languages:
                    c['source'] = u'%%{}\n'.format(language) + c['source']

        nb = new_notebook(cells=cells, metadata=metadata)
        return nb


def _as_dict(metadata):
    if isinstance(metadata, nbformat.NotebookNode):
        return {k: _as_dict(metadata[k]) for k in metadata.keys()}
    return metadata


class RmdWriter(NotebookWriter):

    def __init__(self, markdown=False):
        self.markdown = markdown

    def writes(self, nb):
        default_language = _language(nb.metadata)
        metadata = _as_dict(nb.metadata)

        lines = []
        header_inserted = len(metadata) == 0

        for cell in nb.cells:
            if cell.cell_type == u'raw':
                # Is this the Rmd header? Start and end with '---', and can be parsed with yaml
                if len(lines) == 0 and not header_inserted:
                    header = cell.get(u'source', '').splitlines()
                    if len(header) >= 2 and _header_re.match(header[0]) and _header_re.match(header[-1]):
                        try:
                            header = header[1:-1]
                            yaml.load(u'\n'.join(header))
                            if not self.markdown:
                                header.extend(yaml.safe_dump({u'jupyter': metadata}, default_flow_style=False).splitlines())
                            lines = [u'---'] + header + [u'---']
                            header_inserted = True
                        except yaml.ScannerError:
                            pass
                    if not header_inserted:
                        lines.append(cell.get(u'source', ''))
                else:
                    lines.append(cell.get(u'source', ''))
                if not cell.get(u'metadata', {}).get('noskipline', False):
                    lines.append(u'')
            elif cell.cell_type == u'markdown':
                lines.append(cell.get(u'source', ''))
                if not cell.get(u'metadata', {}).get('noskipline', False):
                    lines.append(u'')
            elif cell.cell_type == u'code':
                input = cell.get(u'source').splitlines()
                cell_metadata = cell.get(u'metadata', {})
                if 'noskipline' in cell_metadata:
                    noskipline = cell_metadata['noskipline']
                    del cell_metadata['noskipline']
                else:
                    noskipline = False
                language = None
                if len(input):
                    for lang, pattern in zip(_jupyter_languages, _jupyter_languages_re):
                        if pattern.match(input[0]):
                            language = lang
                            input = input[1:]
                            break
                if language is None:
                    language = default_language
                if self.markdown:
                    lines.append(u'```' + to_chunk_options(language, cell_metadata))
                else:
                    lines.append(u'```{' + to_chunk_options(language, cell_metadata) + '}')
                if input is not None:
                    lines.extend(input)
                lines.append(u'```')
                if not noskipline:
                    lines.append(u'')

        if not self.markdown and not header_inserted and len(metadata):
            header = yaml.safe_dump({u'jupyter': metadata}, default_flow_style=False).splitlines()
            lines = [u'---'] + header + [u'---', u''] + lines

        lines.append(u'')

        return u'\n'.join(lines)


_reader = RmdReader()
_writer = RmdWriter()

reads = _reader.reads
read = _reader.read
to_notebook = _reader.to_notebook
write = _writer.write
writes = _writer.writes

_md_reader = RmdReader(markdown=True)
_md_writer = RmdWriter(markdown=True)

md_reads = _md_reader.reads
md_read = _md_reader.read
md_to_notebook = _md_reader.to_notebook
md_write = _md_writer.write
md_writes = _md_writer.writes


def readf(nb_file):
    """
    Load the notebook from the desired file
    :param nb_file: file with .ipynb or .Rmd extension
    :return: the notebook
    """
    file, ext = os.path.splitext(nb_file)
    with open(nb_file) as fp:
        if ext == '.Rmd':
            return read(fp)
        elif ext == '.md':
            return md_read(fp)
        elif ext == '.ipynb':
            return nbformat.read(fp, as_version=4)
        else:
            raise TypeError('File {} has incorrect extension (.Rmd or .md or .ipynb expected)'.format(nb_file))


def writef(nb, nb_file):
    """
    Save the notebook in the desired file
    :param nb: notebook
    :param nb_file: file with .ipynb or .Rmd extension
    :return:
    """

    file, ext = os.path.splitext(nb_file)
    with open(nb_file, 'w') as fp:
        if ext == '.Rmd':
            write(nb, fp)
        elif ext == '.md':
            md_write(nb, fp)
        elif ext == '.ipynb':
            nbformat.write(nb, fp)
        else:
            raise TypeError('File {} has incorrect extension (.Rmd or .md or .ipynb expected)'.format(nb_file))


def readme():
    """
    Contents of README.md
    :return:
    """
    readme_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), '..', 'README.md')
    with open(readme_path) as fh:
        return fh.read()
